/* $Header: d:/cvsroot/tads/html/win32/htmlpref.h,v 1.3 1999/05/29 15:51:04 MJRoberts Exp $ */

/* Copyright (c) 1997 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  htmlpref.h - preferences dialog
Function
  
Notes
  
Modified
  10/26/97 MJRoberts  - Creation
*/

#ifndef HTMLPREF_H
#define HTMLPREF_H

#include <windows.h>

#ifndef TADSHTML_H
#include "tadshtml.h"
#endif
#ifndef HTMLPLST_H
#include "htmlplst.h"
#endif


/*
 *   Preference ID's - these are the ID's of the preference settings in
 *   the preference property list.  
 */
enum HTML_pref_id_t
{
    /* first property must start at zero */
    HTML_PREF_PROP_FONT = 0,
    HTML_PREF_MONO_FONT,
    HTML_PREF_FONT_SCALE,
    HTML_PREF_USE_WIN_COLORS,
    HTML_PREF_TEXT_COLOR,
    HTML_PREF_BG_COLOR,
    HTML_PREF_LINK_COLOR,
    HTML_PREF_VLINK_COLOR,
    HTML_PREF_ALINK_COLOR,
    HTML_PREF_UNDERLINE_LINKS,
    HTML_PREF_EMACS_CTRL_V,
    HTML_PREF_ARROW_KEYS_ALWAYS_SCROLL,
    HTML_PREF_ALT_MORE_STYLE,
    HTML_PREF_WINPOS,
    HTML_PREF_HLINK_COLOR,
    HTML_PREF_HOVER_HILITE,
    HTML_PREF_DBGWINPOS,
    HTML_PREF_MUSIC_ON,
    HTML_PREF_SOUNDS_ON,
    HTML_PREF_DBGMAINPOS,
    HTML_PREF_OVERRIDE_COLORS,
    HTML_PREF_LINKS_ON,
    HTML_PREF_LINKS_CTRL,
    HTML_PREF_DIRECTX_HIDEWARN,
    HTML_PREF_DIRECTX_ERROR_CODE,
    HTML_PREF_GRAPHICS_ON,
    HTML_PREF_FILE_SAFETY_LEVEL,
    HTML_PREF_EMACS_ALT_V,
    HTML_PREF_COLOR_STATUS_BG,
    HTML_PREF_COLOR_STATUS_TEXT,
    HTML_PREF_FONT_SERIF,
    HTML_PREF_FONT_SANS,
    HTML_PREF_FONT_SCRIPT,
    HTML_PREF_FONT_TYPEWRITER,
    HTML_PREF_INPFONT_DEFAULT,
    HTML_PREF_INPFONT_NAME,
    HTML_PREF_INPFONT_COLOR,
    HTML_PREF_INPFONT_BOLD,
    HTML_PREF_INPFONT_ITALIC,
    HTML_PREF_MEM_TEXT_LIMIT,
    HTML_PREF_CLOSE_ACTION,
    HTML_PREF_POSTQUIT_ACTION,
    HTML_PREF_INIT_ASK_GAME,
    HTML_PREF_INIT_OPEN_FOLDER,

    /* high water mark - this is always the last property ID */
    HTML_PREF_LAST
};

/*
 *   Main Window Closing action definitions 
 */
const int HTML_PREF_CLOSE_CMD = 1;        /* enter "QUIT" command into game */
const int HTML_PREF_CLOSE_PROMPT = 2;       /* prompt then close the window */
const int HTML_PREF_CLOSE_NOW = 3;               /* close without prompting */

/*
 *   Post-Quit Action definitions
 */
const int HTML_PREF_POSTQUIT_EXIT = 1;   /* wait for a keystroke, then exit */
const int HTML_PREF_POSTQUIT_KEEP = 2;                  /* keep window open */


class CHtmlPreferences
{
public:
    CHtmlPreferences();
    ~CHtmlPreferences();
    
    /* run the dialog */
    void run_preferences_dlg(HWND owner, class CHtmlSysWin_win32 *win);

    /* save/restore the preferences in permanent storage */
    void save();
    void restore();

    /* get my property list */
    class CHtmlPropList *get_proplist() const { return proplist_; }

    /* 
     *   Schedule reformatting of the window.  If 'more_mode' is true,
     *   we'll only reformat the window if it's in MORE mode; otherwise,
     *   we'll unconditionally reformat the window. 
     */
    void schedule_reformat(int more_mode);

    /*
     *   Preference settings 
     */

    /* 
     *   font preferences 
     */
    
    const textchar_t *get_prop_font() const
        { return get_val_str(HTML_PREF_PROP_FONT); }
    void set_prop_font(const textchar_t *newfont)
        { set_val_str(HTML_PREF_PROP_FONT, newfont); }

    const textchar_t *get_mono_font() const
        { return get_val_str(HTML_PREF_MONO_FONT); }
    void set_mono_font(const textchar_t *newfont)
        { set_val_str(HTML_PREF_MONO_FONT, newfont); }
    
    int get_font_scale() const
        { return get_val_longint(HTML_PREF_FONT_SCALE); }
    void set_font_scale(int newscale)
        { set_val_longint(HTML_PREF_FONT_SCALE, newscale); }

    const textchar_t *get_font_serif() const
        { return get_val_str(HTML_PREF_FONT_SERIF); }
    void set_font_serif(const textchar_t *newfont)
        { set_val_str(HTML_PREF_FONT_SERIF, newfont); }

    const textchar_t *get_font_sans() const
        { return get_val_str(HTML_PREF_FONT_SANS); }
    void set_font_sans(const textchar_t *newfont)
        { set_val_str(HTML_PREF_FONT_SANS, newfont); }

    const textchar_t *get_font_script() const
        { return get_val_str(HTML_PREF_FONT_SCRIPT); }
    void set_font_script(const textchar_t *newfont)
        { set_val_str(HTML_PREF_FONT_SCRIPT, newfont); }

    const textchar_t *get_font_typewriter() const
        { return get_val_str(HTML_PREF_FONT_TYPEWRITER); }
    void set_font_typewriter(const textchar_t *newfont)
        { set_val_str(HTML_PREF_FONT_TYPEWRITER, newfont); }

    /*
     *   input font settings 
     */

    const textchar_t *get_inpfont_name() const
        { return get_val_str(HTML_PREF_INPFONT_NAME); }
    void set_inpfont_name(const textchar_t *newfont)
        { set_val_str(HTML_PREF_INPFONT_NAME, newfont); }
    
    HTML_color_t get_inpfont_color() const
        { return get_val_color(HTML_PREF_INPFONT_COLOR); }
    void set_inpfont_color(HTML_color_t newclr)
        { set_val_color(HTML_PREF_INPFONT_COLOR, newclr); }
    
    int get_inpfont_bold() const
        { return get_val_bool(HTML_PREF_INPFONT_BOLD); }
    void set_inpfont_bold(int dflt)
        { set_val_bool(HTML_PREF_INPFONT_BOLD, dflt); }

    int get_inpfont_italic() const
        { return get_val_bool(HTML_PREF_INPFONT_ITALIC); }
    void set_inpfont_italic(int dflt)
        { set_val_bool(HTML_PREF_INPFONT_ITALIC, dflt); }
    

    /* 
     *   color preferences 
     */
    
    int get_use_win_colors() const
        { return get_val_bool(HTML_PREF_USE_WIN_COLORS); }
    void set_use_win_colors(int flag)
        { set_val_bool(HTML_PREF_USE_WIN_COLORS, flag); }

    int get_override_colors() const
        { return get_val_bool(HTML_PREF_OVERRIDE_COLORS); }
    void set_override_colors(int flag)
        { set_val_bool(HTML_PREF_OVERRIDE_COLORS, flag); }

    HTML_color_t get_text_color() const
        { return get_val_color(HTML_PREF_TEXT_COLOR); }
    void set_text_color(HTML_color_t newclr)
        { set_val_color(HTML_PREF_TEXT_COLOR, newclr); }
    
    HTML_color_t get_bg_color() const
        { return get_val_color(HTML_PREF_BG_COLOR); }
    void set_bg_color(HTML_color_t newclr)
        { set_val_color(HTML_PREF_BG_COLOR, newclr); }

    /* 
     *   link preferences 
     */
    
    HTML_color_t get_link_color() const
        { return get_val_color(HTML_PREF_LINK_COLOR); }
    void set_link_color(HTML_color_t newclr)
        { set_val_color(HTML_PREF_LINK_COLOR, newclr); }

    HTML_color_t get_vlink_color() const
        { return get_val_color(HTML_PREF_VLINK_COLOR); }
    void set_vlink_color(HTML_color_t newclr)
        { set_val_color(HTML_PREF_VLINK_COLOR, newclr); }

    HTML_color_t get_hlink_color() const
        { return get_val_color(HTML_PREF_HLINK_COLOR); }
    void set_hlink_color(HTML_color_t newclr)
        { set_val_color(HTML_PREF_HLINK_COLOR, newclr); }

    HTML_color_t get_alink_color() const
        { return get_val_color(HTML_PREF_ALINK_COLOR); }
    void set_alink_color(HTML_color_t newclr)
        { set_val_color(HTML_PREF_ALINK_COLOR, newclr); }
    int get_underline_links() const
        { return get_val_bool(HTML_PREF_UNDERLINE_LINKS); }
    void set_underline_links(int flag)
        { set_val_bool(HTML_PREF_UNDERLINE_LINKS, flag); }

    int get_hover_hilite() const
        { return get_val_bool(HTML_PREF_HOVER_HILITE); }
    void set_hover_hilite(int flag)
        { set_val_bool(HTML_PREF_HOVER_HILITE, flag); }

    /* 
     *   parameterized color settings 
     */

    HTML_color_t get_color_status_bg() const
        { return get_val_color(HTML_PREF_COLOR_STATUS_BG); }
    void set_color_status_bg(HTML_color_t newclr)
        { set_val_color(HTML_PREF_COLOR_STATUS_BG, newclr); }

    HTML_color_t get_color_status_text() const
        { return get_val_color(HTML_PREF_COLOR_STATUS_TEXT); }
    void set_color_status_text(HTML_color_t newclr)
        { set_val_color(HTML_PREF_COLOR_STATUS_TEXT, newclr); }

    /* 
     *   keyboard preferences 
     */

    int get_emacs_ctrl_v() const
        { return get_val_bool(HTML_PREF_EMACS_CTRL_V); }
    void set_emacs_ctrl_v(int flag)
        { set_val_bool(HTML_PREF_EMACS_CTRL_V, flag); }
    int get_emacs_alt_v() const
        { return get_val_bool(HTML_PREF_EMACS_ALT_V); }
    void set_emacs_alt_v(int flag)
        { set_val_bool(HTML_PREF_EMACS_ALT_V, flag); }

    int get_arrow_keys_always_scroll() const
        { return get_val_bool(HTML_PREF_ARROW_KEYS_ALWAYS_SCROLL); }
    void set_arrow_keys_always_scroll(int flag)
        { set_val_bool(HTML_PREF_ARROW_KEYS_ALWAYS_SCROLL, flag); }

    /* 
     *   MORE style 
     */
    
    int get_alt_more_style() const
        { return get_val_bool(HTML_PREF_ALT_MORE_STYLE); }
    void set_alt_more_style(int flag)
        { set_val_bool(HTML_PREF_ALT_MORE_STYLE, flag); }

    /*
     *   Memory text size limit.  A limit of zero indicates that there is
     *   no limit on how large the text buffer can grow; otherwise, the
     *   limit is the maximum number of bytes that we'll allow in the text
     *   array.  
     */
    unsigned long get_mem_text_limit() const
        { return (unsigned long)get_val_longint(HTML_PREF_MEM_TEXT_LIMIT); }
    void set_mem_text_limit(unsigned long val)
        { set_val_longint(HTML_PREF_MEM_TEXT_LIMIT, (long)val); }

    /*
     *   Ask for game at startup.  If this is set, we'll prompt for a game
     *   immediately at startup; otherwise, we'll just stand by until the
     *   user explicitly opens a game.  
     */
    int get_init_ask_game() const
        { return (int)get_val_longint(HTML_PREF_INIT_ASK_GAME); }
    void set_init_ask_game(int val)
        { set_val_longint(HTML_PREF_INIT_ASK_GAME, (long)val); }

    /*
     *   Initial open folder.  This is the folder we'll start in when
     *   asking for a game to open.
     */
    const textchar_t *get_init_open_folder() const
        { return get_val_str(HTML_PREF_INIT_OPEN_FOLDER); }
    void set_init_open_folder(const textchar_t *str)
        { set_val_str(HTML_PREF_INIT_OPEN_FOLDER, str); }
    
    /*
     *   Main Window Close Action - this specifies the response to closing
     *   the main game window.  
     */
    int get_close_action() const
        { return (int)get_val_longint(HTML_PREF_CLOSE_ACTION); }
    void set_close_action(int val)
        { set_val_longint(HTML_PREF_CLOSE_ACTION, (long)val); }

    /*
     *   Post-Quit Action - this specifies what happens after the user
     *   quits the current game
     */
    int get_postquit_action() const
        { return (int)get_val_longint(HTML_PREF_POSTQUIT_ACTION); }
    void set_postquit_action(int val)
        { set_val_longint(HTML_PREF_POSTQUIT_ACTION, (long)val); }

    /* 
     *   window position 
     */
    
    const CHtmlRect *get_win_pos() const
        { return get_val_rect(HTML_PREF_WINPOS); }
    void set_win_pos(const CHtmlRect *rect)
        { set_val_rect(HTML_PREF_WINPOS, rect); }

    /* 
     *   debug log window position 
     */
    
    const CHtmlRect *get_dbgwin_pos() const
        { return get_val_rect(HTML_PREF_DBGWINPOS); }
    void set_dbgwin_pos(const CHtmlRect *rect)
        { set_val_rect(HTML_PREF_DBGWINPOS, rect); }

    /* 
     *   debug main control window position 
     */
    
    const CHtmlRect *get_dbgmain_pos() const
        { return get_val_rect(HTML_PREF_DBGMAINPOS); }
    void set_dbgmain_pos(const CHtmlRect *rect)
        { set_val_rect(HTML_PREF_DBGMAINPOS, rect); }

    /* 
     *   music setting 
     */
    
    int get_music_on() const
        { return get_val_bool(HTML_PREF_MUSIC_ON); }
    void set_music_on(int flag)
        { set_val_bool(HTML_PREF_MUSIC_ON, flag); }

    /* 
     *   sound effects setting 
     */
    
    int get_sounds_on() const
        { return get_val_bool(HTML_PREF_SOUNDS_ON); }
    void set_sounds_on(int flag)
        { set_val_bool(HTML_PREF_SOUNDS_ON, flag); }

    /* 
     *   graphics visibility setting 
     */
    
    int get_graphics_on() const
        { return get_val_bool(HTML_PREF_GRAPHICS_ON); }
    void set_graphics_on(int flag)
        { set_val_bool(HTML_PREF_GRAPHICS_ON, flag); }

    /* 
     *   show links setting 
     */
    
    int get_links_on() const
        { return get_val_bool(HTML_PREF_LINKS_ON); }
    void set_links_on(int flag)
        { set_val_bool(HTML_PREF_LINKS_ON, flag); }

    /* 
     *   show links on control key down setting 
     */
    
    int get_links_ctrl() const
        { return get_val_bool(HTML_PREF_LINKS_CTRL); }
    void set_links_ctrl(int flag)
        { set_val_bool(HTML_PREF_LINKS_CTRL, flag); }

    /* 
     *   hide DirectX warnings on startup 
     */
    
    int get_directx_hidewarn() const
        { return get_val_bool(HTML_PREF_DIRECTX_HIDEWARN); }
    void set_directx_hidewarn(int flag)
        { set_val_bool(HTML_PREF_DIRECTX_HIDEWARN, flag); }

    /* 
     *   DirectX error code from last startup 
     */
    
    enum htmlw32_directx_err_t get_directx_error_code() const
        { return ((htmlw32_directx_err_t)
                  get_val_longint(HTML_PREF_DIRECTX_ERROR_CODE)); }
    void set_directx_error_code(htmlw32_directx_err_t val)
        { set_val_longint(HTML_PREF_DIRECTX_ERROR_CODE, (long)val); }

    /* 
     *   File safety level.  The file safety level is a little different
     *   than other preferences, in that it can have a temporary setting
     *   that is different than the persisently-stored preference setting.
     *   This is used when the player uses a safety-level command line
     *   option when starting the run-time: the command-line setting
     *   overrides the setting from the preferences, but does not get
     *   stored in the preferences.
     *   
     *   set_temp_file_safety_level() overrides the settings in the
     *   preferences.  Calling set_file_safety_level() sets the current
     *   setting as well as the stored setting.  
     */
    int get_file_safety_level() const
    {
        return (temp_file_safety_level_set_
                ? temp_file_safety_level_ 
                : (int)get_val_longint(HTML_PREF_FILE_SAFETY_LEVEL));
    }
    void set_file_safety_level(int level)
    {
        /* set the value in the preferences */
        set_val_longint(HTML_PREF_FILE_SAFETY_LEVEL, level);

        /* from now on, use the preference setting */
        temp_file_safety_level_set_ = FALSE;
    }
    void set_temp_file_safety_level(int level)
    {
        /* remember the value in memory, but not in the saved settings */
        temp_file_safety_level_set_ = TRUE;
        temp_file_safety_level_ = level;
    }

    /*
     *   File safety get/set callbacks for TADS run-time
     */
    static void set_io_safety_level_cb(void *ctx, int level)
    {
        /* 
         *   TADS calls us to set the initial safety level from the
         *   command line - set the command line value as the temporary
         *   setting, overriding any saved preference setting for this
         *   session but not updating the preferences 
         */
        ((CHtmlPreferences *)ctx)->set_temp_file_safety_level(level);
    }
    static int get_io_safety_level_cb(void *ctx)
    {
        return ((CHtmlPreferences *)ctx)->get_file_safety_level();
    }

    /* 
     *   get a property value 
     */
    const textchar_t *get_val_str(HTML_pref_id_t id) const
        { return proplist_->get_val_str(id); }
    const CHtmlRect *get_val_rect(HTML_pref_id_t id) const
        { return proplist_->get_val_rect(id); }
    HTML_color_t get_val_color(HTML_pref_id_t id) const
        { return proplist_->get_val_color(id); }
    long get_val_longint(HTML_pref_id_t id) const
        { return proplist_->get_val_longint(id); }
    int get_val_bool(HTML_pref_id_t id) const
        { return proplist_->get_val_bool(id); }

    /* 
     *   set a property value 
     */
    void set_val_str(HTML_pref_id_t id, const textchar_t *str)
        { proplist_->set_val_str(id, str); }
    void set_val_str(HTML_pref_id_t id, const textchar_t *str, size_t len)
        { proplist_->set_val_str(id, str, len); }
    void set_val_rect(HTML_pref_id_t id, const CHtmlRect *rect)
        { proplist_->set_val_rect(id, rect); }
    void set_val_color(HTML_pref_id_t id, HTML_color_t color)
        { proplist_->set_val_color(id, color); }
    void set_val_longint(HTML_pref_id_t id, long val)
        { proplist_->set_val_longint(id, val); }
    void set_val_bool(HTML_pref_id_t id, int val)
        { proplist_->set_val_bool(id, val); }

    /* get the pointer to the custom colors array */
    COLORREF *get_cust_colors() { return cust_colors_; }

private:
    /* read/write a property from/to the system registry */
    void write_to_registry(HTML_pref_id_t id, HKEY key);
    void read_from_registry(HTML_pref_id_t id, HKEY key);
    
    /* preference property list */
    class CHtmlPropList *proplist_;

    /* owning window */
    class CHtmlSysWin_win32 *win_;

    /* registry primary key name */
    static const textchar_t reg_key_name[];

    /* registry custom color value name */
    static const textchar_t custclr_val_name[];

    /* custom color settings */
    COLORREF cust_colors_[16];

    /* temporary file safety level setting */
    int temp_file_safety_level_;
    int temp_file_safety_level_set_ : 1;
};

#endif /* HTMLPREF_H */

